let N_PARTICLES = 1000;
let particles = [];

let WID = 1500;
let HEI = 1500;

let particleCounter = 0;
let sumFrameDuration = 0
function setup() {
  createCanvas(WID, HEI);

  for(let n = 0; n < N_PARTICLES; n++){
    particles.push(new Particle(new createVector(random(WID), random(HEI))));
  }
  noFill()
  frameRate(60)
}

function draw() {
  background(220);

  if(mouseIsPressed){
    particles.push(new Particle(new createVector(mouseX, mouseY)))
  }

  let start = performance.now();
  for(let p of particles){
    p.checkCollision()
    p.checkEdges();
    p.updateState();
    p.display();
  }
  let end = performance.now();

  sumFrameDuration += end - start

  if(frameCount%60==0){
    console.log('.')
  }

  if(frameCount > 1200){
    console.log(sumFrameDuration/1200)
    noLoop()
  }

}

class Particle {
  constructor(pos) {
    this.pos = pos;
    this.velocity = createVector(random(-1, 1), random(-1, 1));
    this.acceleration = createVector(0, 0);

    this.mass = random(3,10)
    this.radius = this.mass;
    this.maxSpeed = 10;
  }

  updateState(newPos) {
    this.velocity.add(this.acceleration);
    this.velocity.limit(this.maxSpeed); // Limit the particle's speed
    this.pos.add(this.velocity);
  }

  checkEdges() {
    if (this.pos.x - this.radius < 0) {
      this.pos.x = this.radius; // Prevent from leaving the canvas from the left side
      this.velocity.x *= -1;
    } else if (this.pos.x + this.radius > width) {
      this.pos.x = width - this.radius; // Prevent from leaving the canvas from the right side
      this.velocity.x *= -1;
    }

    if (this.pos.y - this.radius < 0) {
      this.pos.y = this.radius; // Prevent from leaving the canvas from the top
      this.velocity.y *= -1;
    } else if (this.pos.y + this.radius > height) {
      this.pos.y = height - this.radius; // Prevent from leaving the canvas from the bottom
      this.velocity.y *= -1;
    }
  }

  checkCollision(){
    for(let other of particles){
      if(this != other){
        let distance = this.pos.dist(other.pos);
        let minDistance = this.radius + other.radius + 1;


        if(distance <= minDistance){
           // Calculate collision response
          let normal = p5.Vector.sub(other.pos, this.pos).normalize();
          let relativeVelocity = p5.Vector.sub(other.velocity, this.velocity);
          let impulse = p5.Vector.mult(normal, 2 * p5.Vector.dot(relativeVelocity, normal) / 2);

          // Apply repulsion force to prevent sticking
          let repulsion = p5.Vector.mult(normal, minDistance - distance);

          // Update velocities
          this.velocity.add(p5.Vector.div(impulse, this.mass));
          other.velocity.sub(p5.Vector.div(impulse, other.mass));

          // Apply repulsion force
          this.pos.sub(p5.Vector.div(repulsion, this.mass));
          other.pos.add(p5.Vector.div(repulsion, other.mass));

        }
      }
    }
  }

  display() {
    ellipse(this.pos.x, this.pos.y, this.radius * 2);
  }
}
